<?php namespace Stage\Archive;

use PDO;
use Exception;
use Stage\Archive\Clip\Clip;
use Stage\Stage;
use Stage\Archive\Video\Video;
use Stage\Archive\Clip\Query as ClipQuery;
use Stage\Archive\Video\Query as VideoQuery;

class Directory
{
    const CREATE = 0;
    const ROOT = 1;

    public static $instanceCounter = 0;

    /**
     * @var
     */
    private $id = self::CREATE;

    /**
     * @var
     */
    public $parentId;

    /**
     * @var
     */
    private $name;

    /**
     * @var
     */
    private $description;

    /**
     * @var
     */
    private $link;

    /**
     * @var
     */
    private $totalItems;

    /**
     * The model's attributes.
     *
     * @var array
     */
    protected $attributes = array('id', 'name','description','link','parentId','totalItems');

    /**
     * The attributes that should be hidden for arrays.
     *
     * @var array
     */
    protected $hidden = array('');

    /**
     * The attributes that should be visible in arrays.
     *
     * @var array
     */
    protected $visible = array();

    public function __construct()
    {
        self::$instanceCounter ++;
    }

    public function getId()
    {
        return $this->id;
    }

    public function getParentId()
    {
        return $this->parentId;
    }

    public function getName()
    {
        return $this->name;
    }

    public function getDescription()
    {
        return $this->description;
    }

    public function getLink()
    {
        return $this->link;
    }

    public function getTotalItems()
    {
        return $this->totalItems;
    }

    private function setId($id)
    {
        $this->id = (int) $id;
    }

    public function setDescription($description)
    {
        $this->description = $description;
    }

    public function setName($name)
    {
        $this->name = $name;
    }

    /*public function setParent($directory)
    {

        $directory = $directory instanceof Directory ? $directory : new Directory();

        // Only set the parent if the parent directory id is different from the instance directory id
        if( $directory->getId() != $this->getId() )
            $this->parent = $directory;
    }*/

    public function setLink($link)
    {
        $this->link = $link;
    }

    public function setTotalItems($total)
    {
        return $this->totalItems = $total;
    }

    /**
     * Find a Directory
     *
     * @param $id
     * @param int $toLevel The subdirectory return deep level
     * @return array|int|void
     * @throws Exception
     */
    public static function find($id, $toLevel = 10000, $name = 'ARCHIVE_VIDEO_DIRECTORY')
    {
        // http://stackoverflow.com/questions/8656089/solution-for-fatal-error-maximum-function-nesting-level-of-100-reached-abor
        if (self::$instanceCounter > $toLevel) {
            return 1;
        }

        try {
            $method = "GET";
            $parameters = [[$id, PDO::PARAM_INT]];
            $data = Stage::esp($parameters, $name, $method);

            if (isset($data[0]['return_code']) && ($data[0]['return_code']) == "-2") {
                throw new Exception("The Directory id $id not exists.", $data[0]['return_code']);
            } elseif (isset($data[0]['id_directory'])) {
                $temp = array_map(function ($directory) {
                    return self::createInstance($directory);
                }, $data);

                return $temp[0];
            }
        } catch (Exception $e) {
            throw $e;
        }
    }

    /**
     * all
     *
     * @param bool|false $raw If $raw is true then return the data directly from the SP
     * @return array
     * @throws Exception
     */
    public static function all($raw = false, $name = 'ARCHIVE_VIDEO_DIRECTORIES')
    {
        try {
            $method = "GET";
            $parameters = [];
            $directories = Stage::esp($parameters, $name, $method);

            return !$raw ? array_map(function ($directory) {
                return self::createInstance($directory);
            }, $directories) : $directories;
        } catch (Exception $e) {
            throw $e;
        }
    }

    /**
     * append
     *
     * Save a Directory
     *
     * @return $this
     * @throws Exception
     */
    private function append(Directory $targetDirectory = null, $isUpdating = false, $name = '')
    {
        $targetIdDirectory = ($targetDirectory instanceof Directory) ? $targetDirectory->getId() : self::ROOT;

        if ($isUpdating) {
            $targetIdDirectory = $targetDirectory->parentId;
        }

        $action = $this->getId() == self::CREATE ? "CREATE" : "UPDATE";

        // In case of the user try to move the folder from ROOT -> ROOT
//        if( !isset($targetDirectory->parentId) ){
//            return $this;
//        }

//        echo $this->getId()."aaa";
//        print_r($targetDirectory);
//        print_r($this);
//        die();

        // pass $targetDirectory = 0 to create a new directory
        // pass $targetDirectory = 1 to move to /root
        // pass $targetDirectory = SOME_ID to move to a specific directory

        $parameters = [ [$this->getId(), PDO::PARAM_INT],
                        [$targetIdDirectory, PDO::PARAM_INT],
                        [$this->getName(), PDO::PARAM_INT],
                        [$this->getDescription(), PDO::PARAM_STR],
                        [$this->getLink(), PDO::PARAM_STR]];

        try {
            $method = "SET";
            $data = Stage::esp($parameters, $name, $method);

            if ($action == "CREATE") {
                $this->setId($data[0]['id_directory']);
                \Stage\Events\Event::fire("archive.directory.created", $this);
            } elseif ($action == "UPDATE") {
                \Stage\Events\Event::fire("archive.directory.updated", $this);
            }

            // Update the parentId
            $this->parentId = $targetIdDirectory;

            return $this;
        } catch (Exception $e) {
            throw $e;
        }
    }

    /**
     * appendTo
     *
     * @param Directory $directory If not passing the directory than is inserted in the root
     * @return $this
     * @throws Exception
     */
    public function appendTo(Directory $targetDirectory = null, $isVideo = true)
    {
        try {
            return $this->append($targetDirectory, false, $isVideo ? 'ARCHIVE_VIDEOS_DIRECTORIES' : 'ARCHIVE_CLIPS_DIRECTORIES');
        } catch (Exception $e) {
            throw $e;
        }
    }

    /**
     * save
     *
     * @return Directory
     * @throws Exception
     */
    public function save($isVideo = true)
    {
        if ($this->getId() == self::ROOT) {
            throw new Exception("You are trying to updated the root(id=1) folder. Root folder cannot be updated.");
        }

        try {
            return $this->append($this, true, $isVideo ? 'ARCHIVE_VIDEOS_DIRECTORIES' : 'ARCHIVE_CLIPS_DIRECTORIES');
        } catch (Exception $e) {
            throw $e;
        }
    }

    /**
     * build
     *
     * Build the tree array of directories
     *
     * @return array
     * @throws Exception
     * @url http://stackoverflow.com/questions/4452472/category-hierarchy-php-mysql
     */
    public static function build($name = 'ARCHIVE_VIDEO_DIRECTORIES')
    {
        $directories = self::all(true, $name);
        $rootId = 1;
        $refs = array();
        $list = array();
        foreach ($directories as $row) {
            $ref = & $refs[$row['id_directory']];

            $ref['id']          = (int) $row['id_directory'];
            $ref['parentId']    = (int) $row['parent_id_directory'];
            $ref['name']        = html_entity_decode($row['name'], ENT_COMPAT, 'UTF-8');
            $ref['description'] = html_entity_decode($row['description'], ENT_COMPAT, 'UTF-8');
            $ref['link']        = $row['link'];
            $ref['files']       = $row['total_items'];

            if ($row['parent_id_directory'] == $rootId) {
                //$list[$row['id_directory']] = & $ref; // Put this if you want the array keys has directory id
                $list[] = & $ref;
            } else {
                // $refs[$row['parent_id_directory']]['children'][$row['id_directory']] = & $ref;   // Put this if you want the array keys has directory id
                $refs[$row['parent_id_directory']]['children'][] = & $ref;
            }
        }

        return $list;

//        Database Table: Items
//        Id	Parent Id	Name
//        1	    0	        North America
//        2	    1	        Canada
//        3	    7	        California
//        4	    0	        Europe
//        5	    0	        Antartica
//        6	    3	        Los Angeles
//        7	    1	        United States
//        8	    7	        Florida
//        9	    3	        Hollywood
//        10	2	        Quebec

//        One Pass Parent-Child Array Structure ( original code )
//        $refs = array();
//        $list = array();
//
//        $sql = "SELECT item_id, parent_id, name FROM items ORDER BY name";
//
//        /** @var $pdo \PDO */
//        $result = $pdo->query($sql);
//
//        foreach ($result as $row)
//        {
//            $ref = & $refs[$row['item_id']];
//
//            $ref['parent_id'] = $row['parent_id'];
//            $ref['name']      = $row['name'];
//
//            if ($row['parent_id'] == 0)
//            {
//                $list[$row['item_id']] = & $ref;
//            }
//            else
//            {
//                $refs[$row['parent_id']]['children'][$row['item_id']] = & $ref;
//            }
//        }
    }

    /**
     * findNode
     *
     * Find the node of some directory id
     * @param $tree
     * @param $idDirectory
     * @return array
     */
    public static function findNode($tree, $idDirectory)
    {
        $temp = [];
        foreach ($tree as $node) {
            if ($node['id'] == $idDirectory) {
                return $node;
            } elseif (isset($node['children'])) {
                $temp = self::findNode($node['children'], $idDirectory);
                if ($temp) {
                    return $temp;
                }
            }
        }

        return $temp;
    }

    /**
     * children
     *
     * @param $tree The parameter $tree must start with index 0
     * @return array
     */
    public static function children(array $tree)
    {
        $result = [];
        $ids = [];
        foreach ($tree as $object) {
            $ids = self::findIds($object, array());
            $result[] = array('id' => $object['id'], 'children' => $ids);
        }

        return sizeof($result) ? $result[0] : $result;
    }

    /**
     * findIds
     *
     * This is an helper method
     * @param $child
     * @param $ids
     * @return array
     * @see http://stackoverflow.com/questions/33800847/get-all-children-of-a-nested-tree-recursively
     */
    private static function findIds($child, $ids)
    {
        if (isset($child['children'])) {
            if (count($child['children']) > 0) {
                foreach ($child['children'] as $ch) {
                    $ids[] = [ 'id'=> $ch['id'],
                               'name' => $ch['name'],
                               'description' => $ch['description'],
                               'link' => $ch['link'] ,
                               'parentId' => $ch['parentId']];

                    $ids = self::findIds($ch, $ids);
                }
            }
        }

        return $ids;
    }

    /**
     * clips
     *
     * @param ClipQuery|null $q
     * @param int $idDirectory
     * @return array
     * @throws Exception
     */
    public static function clips(ClipQuery $q = null, $idDirectory = 0)
    {
        $field          = $q->by;
        $sndField       = $q->sndBy;    // TODO ( codigos 100's, default é 0 )
        $trdField       = $q->trdBy;    // TODO
        $term           = $q->term;
        $filter         = $q->filter;
        $sndFilter      = $q->sndFilter;    // TODO ( value das queries especiais )
        $trdFilter      = $q->trdFilter;    // TODO
        $from           = $q->from;
        $to             = $q->to;
        $limit_sys_a    = $q->fromSize;
        $limit_sys_b    = $q->toSize;
        $orderBy        = $q->orderBy;
        $order          = $q->order;

        try {
            $parameters = [[$idDirectory, PDO::PARAM_INT],
                            [$field, PDO::PARAM_INT],
                            [$sndField, PDO::PARAM_INT],
                            [$trdField, PDO::PARAM_INT],
                            [$term, PDO::PARAM_STR],
                            [$filter, PDO::PARAM_INT],
                            [$sndFilter, PDO::PARAM_INT],
                            [$trdFilter, PDO::PARAM_INT],
                            [$limit_sys_a, PDO::PARAM_INT],
                            [$limit_sys_b, PDO::PARAM_INT],
                            [$from, PDO::PARAM_INT],
                            [$to, PDO::PARAM_INT],
                            [$orderBy, PDO::PARAM_INT],
                            [$order, PDO::PARAM_INT]];

            $name = "ARCHIVE_CLIPS_IN_DIRECTORY";
            $method = "GET";
            $data = Stage::esp($parameters, $name, $method);

            return array_map(function ($video) {
                return Clip::createInstance($video);
            }, $data);
        } catch (Exception $e) {
            throw $e;
        }
    }

    /**
     * clipsCount
     *
     * @param Query|null $q
     * @param int $idDirectory
     * @return mixed
     * @throws Exception
     */
    public static function clipsCount(ClipQuery $q = null, $idDirectory = 0)
    {
        $field          = $q->by;
        $sndField       = $q->sndBy;
        $trdField       = $q->trdBy;
        $term           = $q->term;
        $filter         = $q->filter;
        $sndFilter      = $q->sndFilter;
        $trdFilter      = $q->trdFilter;
        $limit_sys_a    = $q->fromSize;
        $limit_sys_b    = $q->toSize;

        try {
            $parameters = [[$idDirectory, PDO::PARAM_INT],
                            [$field, PDO::PARAM_INT],
                            [$sndField, PDO::PARAM_INT],
                            [$trdField, PDO::PARAM_INT],
                            [$term, PDO::PARAM_STR],
                            [$filter, PDO::PARAM_INT],
                            [$sndFilter, PDO::PARAM_INT],
                            [$trdFilter, PDO::PARAM_INT],
                            [$limit_sys_a, PDO::PARAM_INT],
                            [$limit_sys_b, PDO::PARAM_INT]];

            $name = "CLIPS_IN_DIRECTORY_COUNT";
            $method = "GET";
            $data = Stage::esp($parameters, $name, $method);
            return $data[0]['total'];
        } catch (Exception $e) {
            throw $e;
        }
    }

    /**
     * videos
     *
     * Query videos inside a directory. Pass $idDirectory=0 to query assets inside all directories
     * @return []
     * @throws Exception
     */
    public static function videos(VideoQuery $q = null, $idDirectory = 0)
    {
        $room           = $q->room;
        $idUser         = $q->idUser;
        $field          = $q->by;
        $sndField       = $q->sndBy;
        $trdField       = $q->trdBy;
        $term           = $q->term;
        $filter         = $q->filter;
        $sndFilter      = $q->sndFilter;
        $trdFilter      = $q->trdFilter;
        $from           = $q->from;
        $to             = $q->to;
        $limit_sys_a    = $q->fromSize;
        $limit_sys_b    = $q->toSize;
        $orderBy        = $q->orderBy;
        $order          = $q->order;

        try {
            $parameters = [[$room, PDO::PARAM_STR],
                           [$idUser, PDO::PARAM_INT],
                           [$idDirectory, PDO::PARAM_INT],
                           [$field, PDO::PARAM_INT],
                           [$sndField, PDO::PARAM_INT],
                           [$trdField, PDO::PARAM_INT],
                           [$term, PDO::PARAM_STR],
                           [$filter, PDO::PARAM_INT],
                           [$sndFilter, PDO::PARAM_INT],
                           [$trdFilter, PDO::PARAM_INT],
                           [$limit_sys_a, PDO::PARAM_INT],
                           [$limit_sys_b, PDO::PARAM_INT],
                           [$from, PDO::PARAM_INT],
                           [$to, PDO::PARAM_INT],
                           [$orderBy, PDO::PARAM_INT],
                           [$order, PDO::PARAM_INT]];

            $name = 'ARCHIVE_VIDEOS_DIRECTORY';
            $method = "GET";
            $data = Stage::esp($parameters, $name, $method);

            return array_map(function ($video) {
                return Video::createInstance($video);
            }, $data);
        } catch (Exception $e) {
            throw $e;
        }
    }

    /**
     * videosCount
     *
     * Count videos by query
     * @return []
     * @throws Exception
     */
    public static function videosCount(VideoQuery $q = null, $idDirectory = 0)
    {
        $room           = $q->room;
        $idUser         = $q->idUser;
        $field          = $q->by;
        $sndField       = $q->sndBy;
        $trdField       = $q->trdBy;
        $term           = $q->term;
        $filter         = $q->filter;
        $sndFilter      = $q->sndFilter;
        $trdFilter      = $q->trdFilter;
        $limit_sys_a    = $q->fromSize;
        $limit_sys_b    = $q->toSize;

        try {
            $parameters = [[$room, PDO::PARAM_STR],
                            [$idUser, PDO::PARAM_INT],
                            [$idDirectory, PDO::PARAM_INT],
                            [$field, PDO::PARAM_INT],
                            [$sndField, PDO::PARAM_INT],
                            [$trdField, PDO::PARAM_INT],
                            [$term, PDO::PARAM_STR],
                            [$filter, PDO::PARAM_INT],
                            [$sndFilter, PDO::PARAM_INT],
                            [$trdFilter, PDO::PARAM_INT],
                            [$limit_sys_a, PDO::PARAM_INT],
                            [$limit_sys_b, PDO::PARAM_INT]];

            $name = "ARCHIVE_VIDEOS_COUNT";
            $method = "GET";
            $data = Stage::esp($parameters, $name, $method);
            return $data[0]['total'];
        } catch (Exception $e) {
            throw $e;
        }
    }

    /**
     * delete
     *
     * Delete a Directory
     *
     * @return $this
     * @throws Exception
     */
    public static function delete($idDirectory, $softDelete = true, $deleted = true, $name = '')
    {
        if ($softDelete) {
            $parameters = [ [$idDirectory, PDO::PARAM_INT],
                            [$deleted, PDO::PARAM_BOOL]];
            $method = "SET";
        } else {
            $parameters = [[$idDirectory, PDO::PARAM_INT]];
            $method = "DEL";
        }

        try {
            $data = Stage::esp($parameters, $name, $method);
            if (isset($data[0]['return_code']) && $data[0]['return_code'] == "-2") {
                throw new Exception("The Directory id $idDirectory not exists in database", $data[0]['return_code']);
            }

            return $idDirectory;
        } catch (Exception $e) {
            throw $e;
        }
    }

    /**
     * markedToDelete
     *
     * @return array
     * @throws Exception
     */
    public static function markedToDelete($name = '')
    {
        try {
            $parameters = [];
            $method = "GET";
            $directories = Stage::esp($parameters, $name, $method);

            return array_map(function ($directory) {
                return self::createInstance($directory);
            }, $directories);
        } catch (Exception $e) {
            throw $e;
        }
    }

    /**
     * Convert the model instance to JSON.
     *
     * @param  int  $options
     * @return string
     */
    public function toJson()
    {
        return json_encode($this->toArray());
    }

    /**
     * Convert the model instance to an array.
     *
     * @return array
     */
    public function toArray()
    {
        $attributes = $this->attributesToArray();

        return $attributes;
    }

    /**
     * Convert the model's attributes to an array.
     *
     * @return array
     */
    public function attributesToArray()
    {
        $attributes = $this->getArrayableAttributes();
        $arr = array();

        foreach ($attributes as $attribute) {
            if ($attribute == "parent" && $this->$attribute instanceof Directory) {
                $arr[$attribute] = $this->$attribute->toArray()['id'] == self::CREATE ? "/" : $this->$attribute->toArray();
            } elseif (isset($this->$attribute)) {
                $arr[$attribute] = $attribute == "name" || $attribute == "description" ? html_entity_decode($this->$attribute, ENT_COMPAT, 'UTF-8') : $this->$attribute;
            }
        }

        return $arr;
    }

    /**
     * Get an attribute array of all arrayable attributes.
     *
     * @return array
     */
    protected function getArrayableAttributes()
    {
        return $this->getArrayableItems($this->attributes);
    }

    /**
     * Get an attribute array of all arrayable values.
     *
     * @param  array  $values
     * @return array
     */
    protected function getArrayableItems(array $values)
    {
        if (count($this->visible) > 0) {
            return array_intersect_key($values, array_flip($this->visible));
        }

        return array_flip(array_diff_key(array_flip($values), array_flip($this->hidden)));
    }

    /**
     * createInstance
     *
     * @param $data
     */
    public static function createInstance($directory)
    {
        $directoryInstance = new self();
        $directoryInstance->setId($directory['id_directory']);
        if (isset($directory['parent_id_directory'])) {
            $directoryInstance->parentId = $directory['parent_id_directory'];
        } else { // In case of /root folder
            unset($directoryInstance->parentId);
        }
        /*if( isset($directory['parent_id_directory']) )
        {
            $parentDirectory = $directory['parent_id_directory'] == self::CREATE ? self::ROOT : Directory::find($directory['parent_id_directory']);
            $directoryInstance->setParent($parentDirectory);
        }
        else
        {
            unset($directoryInstance->parent);
        }*/
        $directoryInstance->setName(html_entity_decode($directory['name'], ENT_COMPAT, 'UTF-8'));
        $directoryInstance->setDescription(html_entity_decode($directory['description'], ENT_COMPAT, 'UTF-8'));
        $directoryInstance->setLink($directory['link']);
        $directoryInstance->setTotalItems($directory['total_items']);
        return $directoryInstance;
    }

    public function __toString()
    {
        return "My id is: {$this->getId()}\n";
    }
}
