<?php namespace Stage\Live\Profile;

use PDO;
use DateTime;
use Exception;
use Stage\Live\Stream\Stream;
use Stage\Stage;
use Stage\Helper;
use Stage\Live\Tag\Tag;
use Stage\Live\Compose\Compose;
use Stage\Live\Playlist\Playlist;
use Stage\Live\Profile\L10n\L10n;

class Profile
{

    /**
     * @var
     */
    private $id = -1;

    /**
     * An array of L10n localizations
     * @var array
     */
    private $l10n = [];

    /**
     * @var array
     */
    private $isoLanguages = [];

    /**
     * @var
     */
    private $registerDate;

    /**
     * @var
     */
    private $updateDate;

    /**
     * @var
     */
    private $streams = [];

    /**
     * @var
     */
    private $playlists = [];

    /**
     * @var
     */
    private $state;

    /**
     * The model's attributes.
     *
     * @var array
     */
    protected $attributes = array('id', 'isoLanguages', 'streams', 'playlists', 'state');

    /**
     * The attributes that should be mutated to dates.
     *
     * @var array
     */
    protected $dates = array();

    public function __construct()
    {
    }

    public function getId()
    {
        return $this->id;
    }

    public function getState()
    {
        return $this->state;
    }

    public function getL10n()
    {
        return $this->l10n;
    }

    public function getIsoLanguages()
    {
        return $this->isoLanguages;
    }

    private function setId($id)
    {
        $this->id = (int) $id;
    }

    private function setState($state)
    {
        $this->state = (int) $state;
    }

    private function setIsoLanguages($iso)
    {
        if (!in_array($iso, $this->isoLanguages)) {
            $this->isoLanguages[] = $iso;
        }
    }

    public function setRegisterDate($date)
    {
        // Define how is the date format
        if ($date==null) {
            $this->registerDate = "";
        } else {
            //$date = new DateTime($date);
            $this->registerDate = $date->format(DateTime::ATOM);
        }
    }

    public function setUpdateDate($date)
    {
        // Define how is the date format
        if ($date==null) {
            $this->updateDate = "";
        } else {
            //$date = new DateTime($date);
            $this->updateDate = $date->format(DateTime::ATOM);
        }
    }

    public function discardAllL10ns()
    {
        $this->l10n = [];
    }

    /**
     * addL10n
     *
     * @param L10n $l10n
     * @return L10n
     */
    public function addL10n(L10n $l10n)
    {

        // 1. Check if there are already a L10n added, if so than delete the old and put the new
        $l10ns = $this->getL10n();

        $this->l10n = array_values(array_filter($l10ns, function ($l) use ($l10n) {
            return $l->getIsoLanguage() != $l10n->getIsoLanguage();
        }));

        $this->l10n[] = $l10n;

        if (!in_array($l10n->getIsoLanguage(), $this->getIsoLanguages())) {
            $this->isoLanguages[] = $l10n->getIsoLanguage();
        }

        return $l10n;
    }

    /**
     * sizeof
     *
     * @return array
     */
    public static function sizeof()
    {
        $name = "LIVE_PROFILE_NUM";
        $method = "GET";
        $parameters = [];
        $data = Stage::esp($parameters, $name, $method);
        return $data;
    }

    /**
     * dates
     *
     * @return array
     */
    public static function dates($idQuery = 0)
    {
        $name = "LIVE_PROFILE_DATES";
        $method = "GET";
        $parameters = [[$idQuery, PDO::PARAM_INT]];
        $data = Stage::esp($parameters, $name, $method);
        return $data;
    }

    /**
     * getAllIsoLanguages
     *
     * @return array
     */
    public static function getAllIsoLanguages($idProfile = 0)
    {
        $name = "LIVE_PROFILE_ISOLANGUAGES";
        $method = "GET";
        $parameters = [[$idProfile, PDO::PARAM_INT]];
        $data = Stage::esp($parameters, $name, $method);
        return $data;
    }

    /**
     * find
     *
     * Find a Profile
     *
     * @return $this
     * @throws Exception
     */
    public static function find($id)
    {
        try {
            $name = "LIVE_PROFILE";
            $method = "GET";
            $parameters = [[$id, PDO::PARAM_INT]];
            $data = Stage::esp($parameters, $name, $method);

            if (isset($data[0]['return_code']) && isset($data[0]['return_code']) == "-2") {
                throw new Exception("The Profile id $id not exists", $data[0]['return_code']);
            } else {
                return self::createInstance($data[0]);
            }
        } catch (Exception $e) {
            throw $e;
        }
    }

    /**
     * all
     *
     * Returns all Profiles
     *
     * @return Array
     * @throws Exception
     */
    public static function all()
    {
        $q = new Query();
        $q->from = 0;
        $q->to = -1;

        return self::query($q);
    }

    /**
     * query
     *
     * Search Profiles by Query query
     *
     * @return mixed
     * @throws Exception
     */
    public static function query(Query $q)
    {
        $field      = $q->by;
        $sndField   = $q->sndBy;
        $trdField   = $q->trdBy;
        $query      = $q->term;
        $limit_a    = $q->from;
        $limit_b    = $q->to;
        $filter     = $q->filter;
        $sndFilter  = $q->sndFilter;
        $trdFilter  = $q->trdFilter;
        $isoLanguage = $q->isoLanguage;
        $order      = $q->order;
        $orderBy    = $q->orderBy;

        try {
            $method = "GET";
            $name = "LIVE_PROFILES";
            $parameters = [ [$field, PDO::PARAM_INT],
                [$sndField, PDO::PARAM_INT],
                [$trdField, PDO::PARAM_INT],
                [$query, PDO::PARAM_STR],
                [$filter,PDO::PARAM_STR],
                [$sndFilter, PDO::PARAM_STR],
                [$trdFilter, PDO::PARAM_STR],
                [$isoLanguage, PDO::PARAM_STR],
                [$limit_a, PDO::PARAM_INT],
                [$limit_b, PDO::PARAM_INT],
                [$order, PDO::PARAM_INT],
                [$orderBy, PDO::PARAM_INT]];

            $data = Stage::esp($parameters, $name, $method);

            $res = array_map(function ($profile) {
                return self::createInstance($profile);
            }, $data);

            return $res;
        } catch (Exception $e) {
            throw $e;
        }
    }

    /**
     * queryCount
     *
     * @param \Stage\Video\Profile\Query $q
     * @return mixed
     * @throws Exception
     */
    public static function queryCount(Query $q)
    {
        $field      = $q->by;
        $sndField   = $q->sndBy;
        $trdField   = $q->trdBy;
        $query      = $q->term;
        $filter     = $q->filter;
        $sndFilter  = $q->sndFilter;
        $trdFilter  = $q->trdFilter;
        $isoLanguage = $q->isoLanguage;

        try {
            $method = "GET";
            $name = "LIVE_PROFILES_COUNT";
            $parameters = [ [$field, PDO::PARAM_INT],
                [$sndField, PDO::PARAM_INT],
                [$trdField, PDO::PARAM_INT],
                [$query, PDO::PARAM_STR],
                [$filter,PDO::PARAM_STR],
                [$sndFilter, PDO::PARAM_STR],
                [$trdFilter, PDO::PARAM_STR],
                [$isoLanguage, PDO::PARAM_STR]];

            $data = Stage::esp($parameters, $name, $method);
            return (int) $data[0]['total'];
        } catch (Exception $e) {
            throw $e;
        }
    }

    /**
     * duplicate
     *
     * Returns a new Profile duplicated, with a new Profile id
     *
     * @return $this
     * @throws Exception
     */
    public function duplicate()
    {
        try {
            $parameters = [[$this->getId(), PDO::PARAM_INT]];
            $name = "LIVE_PROFILE_CLONE";
            $method = "SET";
            $data = Stage::esp($parameters, $name, $method);
            return self::createInstance($data[0]);
        } catch (Exception $e) {
            throw $e;
        }
    }

    /**
     * systemTags
     *
     * @param int $field
     * @param int $limit_a In bytes
     * @param int $limit_b In bytes
     * @return array
     */
    public static function systemTags($field = 1, $idQuery = 0, $limit_a = 0, $limit_b = 0)
    {
        $parameters = [[$field, PDO::PARAM_INT],
            [$idQuery, PDO::PARAM_INT],
            [$limit_a, PDO::PARAM_INT],
            [$limit_b, PDO::PARAM_INT]];

        $name = "LIVE_PROFILE_SYSTEM_TAGS";
        $method = "GET";
        $tags = Stage::esp($parameters, $name, $method);

        return $tags;
    }

    /**
     * delete
     *
     * Returns a Profile deleted
     *
     * @return $this
     * @throws Exception
     */
    public function delete()
    {
        try {
            $name = "LIVE_PROFILE";
            $method = "DEL";
            $parameters = [[$this->getId(), PDO::PARAM_INT]];
            Stage::esp($parameters, $name, $method);
            return $this;
        } catch (Exception $e) {
            throw $e;
        }
    }

    /**
     * save
     *
     * Save a Profile
     *
     * @return $this
     * @throws Exception
     */
    public function save()
    {
        $parameters = [];
        $idProfile = $this->getId();

        try {

            // If is not set a idProfile then create a new Profile
            if ($idProfile <= 0) {
                $name = "LIVE_PROFILE";
                $method = "SET";
                $data = Stage::esp($parameters, $name, $method);

                if (isset($data[0]['id_profile']) && $data[0]['id_profile'] >= 0) {
                    $idProfile = $data[0]['id_profile'];
                    $this->setId($idProfile);

                    // Save/Create the L10n
                    $this->saveL10n();
                }
            } else {
                // Update the L10n
                $this->saveL10n();
            }
        } catch (Exception $e) {
            throw $e;
        }

        return $this;
    }

    /**
     * saveL10n
     *
     */
    private function saveL10n()
    {
        // Save the L10n
        $idProfile = $this->getId();

        foreach ($this->getL10n() as $l10n) {
            $l10n->setIdProfile($idProfile);
            $l10n->save();
        }
    }

    /**
     * removeL10n
     *
     * Remove an L10n from a Profile
     * @param $isoLanguage
     * @throws Exception
     */
    public function removeL10n($isoLanguage)
    {
        try {

            // Find the L10n
            $l10n = L10n::find($this->getId(), $isoLanguage);

            // Update the Profile L10n's
            $this->l10n = array_filter($this->getL10n(), function ($l10n) use ($isoLanguage) {
                return $l10n->getIsoLanguage() != trim(strtolower($isoLanguage));
            });

            // Update the property isoLanguages
            $iso = trim(strtolower($isoLanguage));
            $this->isoLanguages = array_values(array_diff($this->getIsoLanguages(), [$iso]));

            // Delete the L10n from database
            return $l10n->delete();
        } catch (Exception $e) {
            throw $e;
        }
    }

    /**
     * streams
     *
     * Get the streams from a Profile
     */
    public function streams()
    {
        $name = "LIVE_PROFILE_STREAMS_RELATED";
        $method = "GET";
        $parameters = [ [$this->getId(), PDO::PARAM_INT]];
        $data = Stage::esp($parameters, $name, $method);

        if (isset($data[0]) && isset($data[0]['return_code']) && $data[0]['return_code'] == "-2") {
            return [];
        }

        return array_filter(array_map(function ($stream) {
            return Stream::find($stream['id_live']);
        }, $data), function ($stream) {
            return $stream != [];
        });
    }

    /**
     * playlists
     *
     * Get the playlits from a profile
     */
    public function playlists()
    {
        $name = "LIVE_PROFILE_PLAYLISTS_RELATED";
        $method = "GET";
        $parameters = [ [$this->getId(), PDO::PARAM_INT]];
        $data = Stage::esp($parameters, $name, $method);

        if (isset($data[0]) && isset($data[0]['return_code']) && $data[0]['return_code'] == "-2") {
            return [];
        }

        return array_map(function ($playlist) {
            return Playlist::createInstance($playlist);
        //return Playlist::find($playlist['id_playlist']);
        }, $data);
    }

    /**
     * compose
     * Get the compose of a video or profile
     */
    public static function compose($idProfile, $idStream = 0)
    {
        $name = "LIVE_COMPOSES_RELATED";
        $method = "GET";
        $parameters = [ [$idProfile, PDO::PARAM_INT],
            [$idStream, PDO::PARAM_INT]];
        $data = Stage::esp($parameters, $name, $method);

        if (isset($data[0]) && isset($data[0]['return_code']) && $data[0]['return_code'] == "-2") {
            throw new Exception("Compose not found with idProfile $idProfile and idVideo $idProfile", $data[0]['return_code']);
        }

        return array_map(function ($compose) {
            return Compose::createInstance([$compose]);
        }, $data);
    }

    /**
     * allTagsRelatedWithProfiles
     *
     * @return array
     */
    public static function allTagsRelatedWithProfiles()
    {
        $name = "LIVE_PROFILE_TAGS_RELATED";
        $method = "GET";

        $idQuery = 0;
        $parameters = [[-1, PDO::PARAM_INT],
            [$idQuery, PDO::PARAM_INT]];
        $tags = Stage::esp($parameters, $name, $method);

        if (isset($tags[0]['return_code'])) {
            return [];
        } else {
            $temp = [];
            foreach ($tags as $tag) {
                $temp[] = Tag::createInstance($tag);
            }

            return $temp;
        }
    }

    /**
     * cloneToVideo
     *
     * Clone a Live Profile to a Video Profile
     *
     * @return mixed
     * @throws Exception
     */
    public function cloneToVideo()
    {
        try {
            $name = "LIVE_PROFILE_CLONE_TO_VIDEO";
            $method = "GET";
            $parameters = [[$this->getId(), PDO::PARAM_INT]];

            $data = Stage::esp($parameters, $name, $method);

            if (isset($data[0]['return_code']) && isset($data[0]['return_code']) == "-2") {
                throw new Exception("The Profile id $this->getId() not exists", $data[0]['return_code']);
            } else {
                return $data;
            }

        } catch (Exception $e) {
            throw $e;
        }
    }

    /**
     * Convert the model instance to JSON.
     *
     * @param  int  $options
     * @return string
     */
    public function toJson()
    {
        return json_encode($this->toArray());
    }

    /**
     * Convert the model instance to an array.
     *
     * @return array
     */
    public function toArray()
    {
        $attributes = $this->attributesToArray();

        foreach ($this->getL10n() as $l10n) {
            $attributes['l10n'][] = $l10n->toArray();
        }

        return $attributes;
    }

    /**
     * Convert the model's attributes to an array.
     *
     * @return array
     */
    public function attributesToArray()
    {
        $attributes = $this->attributes;
        $arr = array();

        foreach ($attributes as $attribute) {
            if ($attribute == 'videos') {
                $arr[$attribute] = [];
                foreach ($this->$attribute as $video) {
                    $arr[$attribute][] = $video->toArray();
                }
            }
            if ($attribute == 'streams') {
                $arr[$attribute] = [];
                foreach ($this->$attribute as $stream) {
                    $arr[$attribute][] = $stream->toArray();
                }
            } elseif ($attribute == 'playlists') {
                $arr[$attribute] = [];
                foreach ($this->$attribute as $playlist) {
                    $arr[$attribute][] = $playlist->toArray();
                }
            } else {
                $arr[$attribute] = $attribute == "title" || $attribute == "description" ? html_entity_decode($this->$attribute, ENT_COMPAT, 'UTF-8') : $this->$attribute;
            }
        }

        return $arr;
    }

    /**
     * createInstance
     *
     * @param $data
     * @return Profile
     */
    public static function createInstance($data)
    {
        if (isset($data[0])) {
            $data = $data[0];
        }
        // $data is an array of rows of posts_l10n
        $p = new self();
        $p->setId($data['id_profile']);
        $p->setState($data['state']);

        if (Helper::isValidDate($data['register_date'], "Y-m-d H:i:s")) {
            $p->setRegisterDate(new DateTime($data['register_date']));
        }

        if (Helper::isValidDate($data['update_date'], "Y-m-d H:i:s")) {
            $p->setUpdateDate(new DateTime($data['update_date']));
        }

        $languages = self::getAllIsoLanguages($data['id_profile']);
        foreach ($languages as $l) {
            $iso = $l['iso_language'];

            $p->setIsoLanguages($iso);

            $l10n = L10n::find($p->getId(), $iso);

            // Add the L10n to the Post
            $p->addL10n($l10n);
        }

//        foreach($data as $l)
//        {
//            $l10n = L10n::find($p->getId(), );
//
//            // Add the L10n to the Post
//            $p->addL10n($l10n);
//        }

        $streams = $p->streams();

        foreach ($streams as $stream) {
            $p->streams[] = $stream;
        }

        $playlists = $p->playlists();
        foreach ($playlists as $playlist) {
            $p->playlists[] = $playlist;
        }

        return $p;
    }

    public function __toString()
    {
        return "My id is: {$this->getId()}\n";
    }
}
